/*
Copyright (c) 2010, Geomatics and Cartographic Research Centre, Carleton 
University
All rights reserved.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions are met:

 - Redistributions of source code must retain the above copyright notice, 
   this list of conditions and the following disclaimer.
 - Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.
 - Neither the name of the Geomatics and Cartographic Research Centre, 
   Carleton University nor the names of its contributors may be used to 
   endorse or promote products derived from this software without specific 
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.

$Id$
*/
package ca.carleton.gcrc.olkit.multimedia.ffmpeg;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class FFmpeg {
	
	static private Pattern patternInput0 = Pattern.compile("^Input #0,\\s*([^,]*),");
	static private Pattern patternDuration = Pattern.compile("^\\s*Duration:\\s*((\\d+):(\\d+):(\\d+\\.\\d+)),.*start:\\s*(\\d+\\.\\d+),.*bitrate:\\s*((\\d+)\\s*(.*))");
	static private Pattern patternVideo = Pattern.compile("^\\s*Stream.*Video:\\s*([^,]*),.*,\\s*(\\d+)x(\\d+)");
	static private Pattern patternAudio = Pattern.compile("^\\s*Stream.*Audio:\\s*([^,]*)");
	static private Pattern patternTime = Pattern.compile("^\\s*frame=.*time=\\s*(\\d+\\.\\d*)");

	static private FFmpegInfo availability;
	synchronized static public FFmpegInfo getInfo() {
		if( null == availability ) {
			FFmpegInfo avail = new FFmpegInfo();
			
			Runtime rt = Runtime.getRuntime();
			try {
				Process p = rt.exec("ffmpeg -version", null, null);
				InputStream is = p.getInputStream();
				InputStreamReader isr = new InputStreamReader(is);
				BufferedReader bufReader = new BufferedReader(isr);
				String line = bufReader.readLine();
				avail.isAvailable = true;
				avail.longVersion = line;
			} catch (IOException e) {
				avail.isAvailable = false;
			}
			
			availability = avail;
		}
		return availability;
	}
	
	private FFmpegProgress ffmpegProgress;
	
	public FFmpeg() {
		
	}
	
	public FFmpeg(FFmpegProgress ffmpegProgress) {
		this.ffmpegProgress = ffmpegProgress;
	}
	
	public VideoInfo getVideoInfo(File videoFile) throws Exception {
		VideoInfo info = new VideoInfo();
		info.file = videoFile;
		
		Runtime rt = Runtime.getRuntime();
		try {
			String command = "ffmpeg -i "+videoFile.getAbsolutePath();
			Process p = rt.exec(command, null, null);
			InputStream is = p.getErrorStream();
			InputStreamReader isr = new InputStreamReader(is);
			BufferedReader bufReader = new BufferedReader(isr);
			String line = bufReader.readLine();
			boolean inputFound = false;
			while( null != line ) {
				// Ignore lines until we find the start of the movie description 'Input #0'
				if( false == inputFound ) {
					Matcher matcher = patternInput0.matcher(line);
					if( matcher.find() ) {
						inputFound = true;
						info.fileType = matcher.group(1).trim();
					}
				} else {
					Matcher matcherDuration = patternDuration.matcher(line);
					Matcher matcherVideo = patternVideo.matcher(line);
					Matcher matcherAudio = patternAudio.matcher(line);

					if( matcherDuration.find() ) {
						info.durationString = matcherDuration.group(1);
						int hours = Integer.parseInt( matcherDuration.group(2) );
						int mins = Integer.parseInt( matcherDuration.group(3) );
						float secs = Float.parseFloat( matcherDuration.group(4) );
						info.durationInSec = new Float( secs + (60.0*mins) + (3600.0*hours));
						info.startString = matcherDuration.group(5);
						info.startInSec = Float.parseFloat( info.startString );
						info.bitRateString = matcherDuration.group(6);
						int bitRate = Integer.parseInt( matcherDuration.group(7) );
						String bitRateUnit = matcherDuration.group(8).trim();
						if( "kb/s".equals(bitRateUnit) ) {
							info.bitRate = new Integer( bitRate * 1000 );
						} else if( "b/s".equals(bitRateUnit) ) {
							info.bitRate = new Integer( bitRate );
						} else if( "Mb/s".equalsIgnoreCase(bitRateUnit) ) {
							info.bitRate = new Integer( bitRate * 1000000 );
						}
						
					} else if( matcherVideo.find() ) {
						info.videoCodec = matcherVideo.group(1);
						info.width = new Integer( Integer.parseInt( matcherVideo.group(2) ) );
						info.height = new Integer( Integer.parseInt( matcherVideo.group(3) ) );
						
					} else if( matcherAudio.find() ) {
						info.audioCodec = matcherAudio.group(1);
					}
					
				}
				
				line = bufReader.readLine();
			}
		} catch (IOException e) {
			throw new Exception("Error while reading info about "+videoFile.getAbsolutePath(),e);
		}
		
		return info;
	}

	public void convertVideo(File inputFile, File outputFile) throws Exception {
		VideoInfo inputVideo = getVideoInfo(inputFile);
		convertVideo(inputVideo, outputFile);
	}
	
	public void convertVideo(VideoInfo inputVideo, File outputFile) throws Exception {
		
		Runtime rt = Runtime.getRuntime();
		try {
			String command = "ffmpeg -i "+inputVideo.file.getAbsolutePath()+" -y -acodec libfaac -ab 48000 -ac 2 -vcodec libx264 -b 128000 -s 320x240 -threads 0 -f mp4 "+outputFile.getAbsolutePath();
			Process p = rt.exec(command, null, null);
			InputStream is = p.getErrorStream();
			InputStreamReader isr = new InputStreamReader(is);
			BufferedReader bufReader = new BufferedReader(isr);
			String line = bufReader.readLine();
			while( null != line ) {

				if( null != inputVideo.durationInSec && null != ffmpegProgress ) {
					Matcher matcher = patternTime.matcher(line);
					if( matcher.find() ) {
						float time = Float.parseFloat( matcher.group(1) );
						float percent = time / inputVideo.durationInSec.floatValue();
						int percentInt = (int)(percent * 100);
						if( percentInt > 100 ) {
							percentInt = 100;
						}
						ffmpegProgress.updateProgress(percentInt);
					}
				}
				
				line = bufReader.readLine();
			}
		} catch (IOException e) {
			throw new Exception("Error while reading converting "+inputVideo.file.getAbsolutePath(),e);
		}
	}

	public void convertAudio(File inputFile, File outputFile) throws Exception {
		VideoInfo inputVideo = getVideoInfo(inputFile);
		convertAudio(inputVideo, outputFile);
	}
	
	public void convertAudio(VideoInfo inputVideo, File outputFile) throws Exception {
		
		Runtime rt = Runtime.getRuntime();
		try {
			String command = "ffmpeg -i "+inputVideo.file.getAbsolutePath()+" -y -acodec libfaac -ab 48000 -ac 2 -threads 0 -f mp4 "+outputFile.getAbsolutePath();
			Process p = rt.exec(command, null, null);
			InputStream is = p.getErrorStream();
			InputStreamReader isr = new InputStreamReader(is);
			BufferedReader bufReader = new BufferedReader(isr);
			String line = bufReader.readLine();
			while( null != line ) {

				if( null != inputVideo.durationInSec && null != ffmpegProgress ) {
					Matcher matcher = patternTime.matcher(line);
					if( matcher.find() ) {
						float time = Float.parseFloat( matcher.group(1) );
						float percent = time / inputVideo.durationInSec.floatValue();
						int percentInt = (int)(percent * 100);
						if( percentInt > 100 ) {
							percentInt = 100;
						}
						ffmpegProgress.updateProgress(percentInt);
					}
				}
				
				line = bufReader.readLine();
			}
		} catch (IOException e) {
			throw new Exception("Error while reading converting "+inputVideo.file.getAbsolutePath(),e);
		}
	}
	
	public void createThumbnail(File inputFile, File outputFile) throws Exception {
		VideoInfo inputVideo = getVideoInfo(inputFile);
		createThumbnail(inputVideo, outputFile);
	}
	
	public void createThumbnail(VideoInfo inputVideo, File outputFile) throws Exception {
		
		Runtime rt = Runtime.getRuntime();
		try {
			String command = "ffmpeg -i "+inputVideo.file.getAbsolutePath()+" -y -ss 00:00:05 -s 80x60 -r 1 -vframes 1 -f image2 "+outputFile.getAbsolutePath();
			rt.exec(command, null, null);
		} catch (IOException e) {
			throw new Exception("Error while creating thumbnail for "+inputVideo.file.getAbsolutePath(),e);
		}
	}
}
