/*
Copyright (c) 2010, Geomatics and Cartographic Research Centre, Carleton 
University
All rights reserved.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions are met:

 - Redistributions of source code must retain the above copyright notice, 
   this list of conditions and the following disclaimer.
 - Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.
 - Neither the name of the Geomatics and Cartographic Research Centre, 
   Carleton University nor the names of its contributors may be used to 
   endorse or promote products derived from this software without specific 
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.

$Id$
*/
package ca.carleton.gcrc.olkit.multimedia.imageMagick;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class ImageMagick {

	static private Pattern patternInfoGeometry = Pattern.compile("^\\s*Geometry:\\s*(\\d+)x(\\d+)");
	static private Pattern patternInfoFormat = Pattern.compile("^\\s*Format:\\s*([^\\s]+)");
	static private Pattern patternProgressLoad = Pattern.compile("^\\s*Load\\s+image:\\s*(\\d+)%");
	static private Pattern patternProgressResize = Pattern.compile("^\\s*Resize\\s+image:\\s*(\\d+)%");
	static private Pattern patternProgressSave = Pattern.compile("^\\s*Save\\s+image:\\s*(\\d+)%");
	
	static private ImageMagickInfo availability;
	synchronized static public ImageMagickInfo getInfo() {
		if( null == availability ) {
			ImageMagickInfo avail = new ImageMagickInfo();
			
			avail.isAvailable = true;
			
			// Identify
			Runtime rt = Runtime.getRuntime();
			try {
				Process p = rt.exec("identify -version", null, null);
				InputStream is = p.getInputStream();
				InputStreamReader isr = new InputStreamReader(is);
				BufferedReader bufReader = new BufferedReader(isr);
				String line = bufReader.readLine();
				avail.identifyVersion = line;
				avail.isIdentifyAvailable = true;
			} catch (IOException e) {
				avail.isIdentifyAvailable = false;
				avail.isAvailable = false;
			}
			
			// Convert
			try {
				Process p = rt.exec("convert -version", null, null);
				InputStream is = p.getInputStream();
				InputStreamReader isr = new InputStreamReader(is);
				BufferedReader bufReader = new BufferedReader(isr);
				String line = bufReader.readLine();
				avail.convertVersion = line;
				avail.isConvertAvailable = true;
			} catch (IOException e) {
				avail.isConvertAvailable = false;
				avail.isAvailable = false;
			}
			
			availability = avail;
		}
		return availability;
	}

	private ImageMagickProgress imageMagickProgress;
	
	public ImageMagick() {
		
	}
	
	public ImageMagick(ImageMagickProgress imageMagickProgress) {
		this.imageMagickProgress = imageMagickProgress;
	}
	
	public ImageInfo getImageInfo(File imageFile) throws Exception {
		ImageInfo info = new ImageInfo();
		info.file = imageFile;
		
		Runtime rt = Runtime.getRuntime();
		try {
			String command = "identify -verbose "+imageFile.getAbsolutePath();
			Process p = rt.exec(command, null, null);
			InputStream is = p.getInputStream();
			InputStreamReader isr = new InputStreamReader(is);
			BufferedReader bufReader = new BufferedReader(isr);
			String line = bufReader.readLine();
			while( null != line ) {

				Matcher matcherGeometry = patternInfoGeometry.matcher(line);
				Matcher matcherFormat = patternInfoFormat.matcher(line);
				
				if( matcherGeometry.find() ) {
					info.width = Integer.parseInt( matcherGeometry.group(1) );
					info.height = Integer.parseInt( matcherGeometry.group(2) );
					
				} else if( matcherFormat.find() ) {
					info.format = matcherFormat.group(1);
				}
				
				line = bufReader.readLine();
			}
		} catch (IOException e) {
			throw new Exception("Unabel to determine image info",e);
		}
		
		return info;
	}
	
	public void convertImage(File imageFile, File outputFile) throws Exception {
		ImageInfo imageInfo = getImageInfo(imageFile);
		convertImage(imageInfo, outputFile);
	}
	
	public void convertImage(ImageInfo imageInfo, File outputFile) throws Exception {
		
		Runtime rt = Runtime.getRuntime();
		try {
			String monitorString = "";
			if( null != imageMagickProgress ) {
				monitorString = "-monitor ";
			}
			String command = "convert "+monitorString+imageInfo.file.getAbsolutePath()+" -compress JPEG -quality 70 "+outputFile.getAbsolutePath();
			Process p = rt.exec(command, null, null);
			InputStream is = p.getInputStream();
			InputStreamReader isr = new InputStreamReader(is);
			BufferedReader bufReader = new BufferedReader(isr);
			String line = bufReader.readLine();
			while( null != line ) {
				line = line.trim();

				if( null != imageMagickProgress ) {
					Matcher matcherLoad = patternProgressLoad.matcher(line);
					Matcher matcherResize = patternProgressResize.matcher(line);
					Matcher matcherSave = patternProgressSave.matcher(line);
					
					if( matcherLoad.find() ) {
						int value = Integer.parseInt( matcherLoad.group(1) );
						int percent = value / 300;
						imageMagickProgress.updateProgress(percent);
						
					} else if( matcherResize.find() ) {
						int value = Integer.parseInt( matcherResize.group(1) );
						int percent = (value+100) / 300;
						imageMagickProgress.updateProgress(percent);
						
					} else if( matcherSave.find() ) {
						int value = Integer.parseInt( matcherSave.group(1) );
						int percent = (value+200) / 300;
						imageMagickProgress.updateProgress(percent);
					}
				}
				
				line = bufReader.readLine();
			}
		} catch (IOException e) {
			throw new Exception("Error while converting image "+imageInfo.file.getAbsolutePath(),e);
		}
	}

	public void resizeImage(File imageFile, File outputFile, int maxWidth, int maxHeight) throws Exception {
		ImageInfo imageInfo = getImageInfo(imageFile);
		resizeImage(imageInfo, outputFile, maxWidth, maxHeight);
	}
	
	public void resizeImage(ImageInfo imageInfo, File outputFile, int maxWidth, int maxHeight) throws Exception {
		
		Runtime rt = Runtime.getRuntime();
		try {
			String monitorString = "";
			if( null != imageMagickProgress ) {
				monitorString = "-monitor ";
			}
			String command = "convert "+monitorString+imageInfo.file.getAbsolutePath()+" -resize "+maxWidth+"x"+maxHeight+" -compress JPEG -quality 70 "+outputFile.getAbsolutePath();
			Process p = rt.exec(command, null, null);
			InputStream is = p.getInputStream();
			InputStreamReader isr = new InputStreamReader(is);
			BufferedReader bufReader = new BufferedReader(isr);
			String line = bufReader.readLine();
			while( null != line ) {
				line = line.trim();

				if( null != imageMagickProgress ) {
					Matcher matcherLoad = patternProgressLoad.matcher(line);
					Matcher matcherResize = patternProgressResize.matcher(line);
					Matcher matcherSave = patternProgressSave.matcher(line);
					
					if( matcherLoad.find() ) {
						int value = Integer.parseInt( matcherLoad.group(1) );
						int percent = value / 300;
						imageMagickProgress.updateProgress(percent);
						
					} else if( matcherResize.find() ) {
						int value = Integer.parseInt( matcherResize.group(1) );
						int percent = (value+100) / 300;
						imageMagickProgress.updateProgress(percent);
						
					} else if( matcherSave.find() ) {
						int value = Integer.parseInt( matcherSave.group(1) );
						int percent = (value+200) / 300;
						imageMagickProgress.updateProgress(percent);
					}
				}
				
				line = bufReader.readLine();
			}
		} catch (IOException e) {
			throw new Exception("Error while converting image "+imageInfo.file.getAbsolutePath(),e);
		}
	}
}
